/*
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package com.inteligr8.alfresco.activiti.share;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import org.activiti.engine.ProcessEngine;
import org.activiti.engine.repository.ModelQuery;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.domain.PageRequest;
import org.springframework.data.domain.Pageable;
import org.springframework.stereotype.Component;

import com.activiti.domain.editor.AbstractModel;
import com.activiti.domain.editor.Model;
import com.activiti.domain.editor.ModelShareInfo;
import com.activiti.domain.editor.SharePermission;
import com.activiti.domain.idm.Group;
import com.activiti.repository.editor.ModelShareInfoRepository;
import com.activiti.service.api.GroupService;
import com.activiti.service.api.ModelService;
import com.activiti.service.editor.ModelShareService;

@Component
public class ModelShareWorker implements Bootstrappable {
    
    private final Logger logger = LoggerFactory.getLogger(this.getClass());
    
    @Value("${inteligr8.modelShareExtension.modelChunkSize:50}")
    private int modelChunkSize;
    
    @Value("${inteligr8.modelShareExtension.shareChunkSize:25}")
    private int shareChunkSize;
    
    @Value("${inteligr8.modelShareExtension.groups.canRead:#{null}}")
    private String readGroupNames;
    
    @Value("${inteligr8.modelShareExtension.groups.canWrite:#{null}}")
    private String writeGroupNames;
    
    @Value("${inteligr8.modelShareExtension.groups.canReadApps:#{null}}")
    private String readAppDefsGroupNames;
    
    @Value("${inteligr8.modelShareExtension.groups.canWriteApps:#{null}}")
    private String writeAppDefsGroupNames;
    
    @Value("${inteligr8.modelShareExtension.groups.canReadProcesses:#{null}}")
    private String readProcessDefsGroupNames;
    
    @Value("${inteligr8.modelShareExtension.groups.canWriteProcesses:#{null}}")
    private String writeProcessDefsGroupNames;
    
    @Value("${inteligr8.modelShareExtension.groups.canReadForms:#{null}}")
    private String readFormDefsGroupNames;
    
    @Value("${inteligr8.modelShareExtension.groups.canWriteForms:#{null}}")
    private String writeFormDefsGroupNames;
    
    @Autowired
    private ProcessEngine services;
    
    @Autowired
    private ModelShareService shareService;
    
    @Autowired
    private ModelShareInfoRepository shareInfoRepo;
    
    @Autowired
    private ModelService modelService;
    
    @Autowired
    private GroupService groupService;
    
    @Autowired
    private TenantFinderService tenantService;
    
    private Long tenantId;
    private Map<HashableGroup, SharePermission> groups;
    private Map<HashableGroup, SharePermission> appDefsGroups;
    private Map<HashableGroup, SharePermission> processDefsGroups;
    private Map<HashableGroup, SharePermission> formDefsGroups;
    
    @Override
    public void onBootstrap() {
        this.logger.trace("onBootstrap()");

        this.tenantId = this.tenantService.findTenantId();
        this.groups = this.toMap(this.findGroups(this.readGroupNames), SharePermission.READ);
        this.toMap(this.groups, this.findGroups(this.writeGroupNames), SharePermission.WRITE);
        this.appDefsGroups = this.toMap(this.findGroups(this.readAppDefsGroupNames), SharePermission.READ);
        this.toMap(this.appDefsGroups, this.findGroups(this.writeAppDefsGroupNames), SharePermission.WRITE);
        this.processDefsGroups = this.toMap(this.findGroups(this.readProcessDefsGroupNames), SharePermission.READ);
        this.toMap(this.processDefsGroups, this.findGroups(this.writeProcessDefsGroupNames), SharePermission.WRITE);
        this.formDefsGroups = this.toMap(this.findGroups(this.readFormDefsGroupNames), SharePermission.READ);
        this.toMap(this.formDefsGroups, this.findGroups(this.writeFormDefsGroupNames), SharePermission.WRITE);

        this.logger.info("Will share all models with groups: {}", this.groups);
        this.logger.info("Will share app models with groups: {}", this.appDefsGroups);
        this.logger.info("Will share process models with groups: {}", this.processDefsGroups);
        this.logger.info("Will share form models with groups: {}", this.formDefsGroups);
    }
    
    private Set<HashableGroup> findGroups(String commaDelimitedGroups) {
        commaDelimitedGroups = StringUtils.trimToNull(commaDelimitedGroups);
        if (commaDelimitedGroups == null)
            return Collections.emptySet();
            
        Set<HashableGroup> groups = new HashSet<>();
        
        for (String groupName : commaDelimitedGroups.split(",")) {
            groupName = groupName.trim();
            if (groupName.length() == 0)
                continue;
            
            String[] groupParts = groupName.split(":");
            String prefix = groupParts.length > 1 ? groupParts[0] : null;
            if (prefix != null)
                groupName = groupParts[1];

            if (prefix == null) {
                List<Group> tenantGroups = this.groupService.getGroupByNameAndTenantId(groupName, this.tenantId);
                if (!tenantGroups.isEmpty()) {
                    this.logger.debug("Found groups: {}: {}", groupName, this.getIds(tenantGroups));
                    for (Group group : tenantGroups)
                        groups.add(new HashableGroup(group));
                } else {
                    Group group = this.groupService.getGroupByExternalIdAndTenantId(groupName, this.tenantId);
                    if (group != null) {
                        this.logger.debug("Found group: {}: {}", groupName, group.getId());
                        groups.add(new HashableGroup(group));
                    } else {
                        this.logger.warn("Could not find group; cannot use for sharing models: {}", groupName);
                    }
                }
            } else if (prefix.startsWith("e")) {
                Group group = this.groupService.getGroupByExternalIdAndTenantId(groupName, this.tenantId);
                if (group != null) {
                    this.logger.debug("Found group: {}: {}", groupName, group.getId());
                    groups.add(new HashableGroup(group));
                } else {
                    this.logger.warn("Could not find group; cannot use for sharing models: {}", groupName);
                }
            } else if (prefix.startsWith("o")) {
                List<Group> tenantGroups = this.groupService.getFunctionalGroupsByNameAndTenantId(groupName, this.tenantId);
                if (!tenantGroups.isEmpty()) {
                    this.logger.debug("Found functional groups: {}: {}", groupName, this.getIds(tenantGroups));
                    for (Group group : tenantGroups)
                        groups.add(new HashableGroup(group));
                } else {
                    this.logger.warn("Could not find group; cannot use for sharing models: {}", groupName);
                }
            } else if (prefix.startsWith("s")) {
                List<Group> tenantGroups = this.groupService.getSystemGroupWithName(groupName, this.tenantId);
                if (!tenantGroups.isEmpty()) {
                    this.logger.debug("Found system groups: {}: {}", groupName, this.getIds(tenantGroups));
                    for (Group group : tenantGroups)
                        groups.add(new HashableGroup(group));
                } else {
                    this.logger.warn("Could not find group; cannot use for sharing models: {}", groupName);
                }
            } else {
                this.logger.warn("Prefix not recognized; cannot use for sharing models: {}:{}", prefix, groupName);
            }
        }
        
        return groups;
    }
    
    private Collection<Long> getIds(Collection<Group> groups) {
        List<Long> ids = new ArrayList<>(groups.size());
        for (Group group : groups)
            ids.add(group.getId());
        return  ids;
    }
    
    private <K, V> Map<K, V> toMap(Collection<K> keys, V value) {
        Map<K, V> map = new HashMap<>();
        this.toMap(map, keys, value);
        return map;
    }
    
    private <K, V> void toMap(Map<K, V> map, Collection<K> keys, V value) {
        for (K key : keys)
            map.put(key, value);
    }
    
    public void shareAllModels() {
        this.shareApsFormModels();
        this.shareApsDataModels();
        this.shareApsDecisionTableModels();
        this.shareApsProcessModels();
        this.shareApsAppModels();
        
        if (this.tenantId != null) {
            this.shareAllActivitiModels(this.tenantId);
        } else {
            this.shareAllActivitiModels(null);
        }
    }
    
    private void shareApsFormModels() {
        this.logger.trace("Discovering APS form models");
        List<AbstractModel> models = this.modelService.getModelsByModelTypeAndReferenceId(Model.MODEL_TYPE_FORM, null);
        for (AbstractModel model : models) {
            this.logger.debug("Discovered APS form model: {} [{}]", model.getName(), model.getId());
            this.shareModel((Model) model);
        }
    }
    
    private void shareApsDataModels() {
        this.logger.trace("Discovering APS data models");
        List<AbstractModel> models = this.modelService.getModelsByModelTypeAndReferenceId(Model.MODEL_TYPE_DATA_MODEL, null);
        for (AbstractModel model : models) {
            this.logger.debug("Discovered APS data model: {} [{}]", model.getName(), model.getId());
            this.shareModel((Model) model);
        }
    }
    
    private void shareApsDecisionTableModels() {
        this.logger.trace("Discovering APS decision table models");
        List<AbstractModel> models = this.modelService.getModelsByModelTypeAndReferenceId(Model.MODEL_TYPE_DECISION_TABLE, null);
        for (AbstractModel model : models) {
            this.logger.debug("Discovered APS decision table model: {} [{}]", model.getName(), model.getId());
            this.shareModel((Model) model);
        }
    }
    
    private void shareApsProcessModels() {
        this.logger.trace("Discovering APS process models");
        List<AbstractModel> models = this.modelService.getModelsByModelTypeAndReferenceId(Model.MODEL_TYPE_BPMN, null);
        for (AbstractModel model : models) {
            this.logger.debug("Discovered APS process model: {} [{}]", model.getName(), model.getId());
            this.shareModel((Model) model);
        }
    }
    
    private void shareApsAppModels() {
        this.logger.trace("Discovering APS app models");
        List<AbstractModel> models = this.modelService.getModelsByModelTypeAndReferenceId(Model.MODEL_TYPE_APP, null);
        for (AbstractModel model : models) {
            this.logger.debug("Discovered APS app model: {} [{}]", model.getName(), model.getId());
            this.shareModel((Model) model);
        }
    }
    
    private void shareAllActivitiModels(Long tenantId) {
        this.logger.trace("Discovering activiti models in tenant {} ...", tenantId);
        
        long modelCount = 0L;
        int page = 1;
        int perPage = this.modelChunkSize;
        
        ModelQuery query = this.services.getRepositoryService().createModelQuery()
        		.latestVersion();
        if (tenantId != null)
            query.modelTenantId(String.valueOf(tenantId));
        else query.modelWithoutTenantId();

        this.logger.trace("Discovering activiti models with page size: {}", perPage);
        List<org.activiti.engine.repository.Model> models = query.listPage((page-1)*perPage, perPage);
        if (models.isEmpty()) {
        	this.logger.debug("No activiti models; procDef count: {}", this.services.getRepositoryService().createProcessDefinitionQuery().latestVersion().count());
        }
        
        while (!models.isEmpty()) {
            this.logger.trace("Fetched page #{} of {} activiti models in tenant {}", page, models.size(), tenantId);
            
            for (org.activiti.engine.repository.Model model : models) {
                this.logger.debug("Discovered model: {}: {}: {}", model.getCategory(), model.getId(), model.getName());
                this.shareModel(model);
            }
            
            modelCount += models.size();
            page++;
            models = query.listPage((page-1)*perPage, perPage);
        }

        this.logger.trace("Discovered {} activiti models in tenant {}", modelCount, tenantId);
    }
    
    public void shareModel(org.activiti.engine.repository.Model orgModel) {
        Model model = (Model) this.modelService.getModel(Long.valueOf(orgModel.getId()));
        this.shareModel(model);
    }
    
    public void shareModel(Model model) {
        Map<HashableGroup, SharePermission> shares = this.fetchCurrentModelShares(model);

        switch (model.getModelType()) {
            case Model.MODEL_TYPE_APP:
                for (Entry<HashableGroup, SharePermission> group : this.groups.entrySet()) {
                    if (this.appDefsGroups.containsKey(group.getKey())) {
                        this.logger.trace("The default group permission is overridden by the app group permission: {}", group.getKey());
                    } else {
                        this.share(shares, model, group.getKey(), group.getValue());
                    }
                }
                for (Entry<HashableGroup, SharePermission> group : this.appDefsGroups.entrySet())
                    this.share(shares, model, group.getKey(), group.getValue());
                break;
            case Model.MODEL_TYPE_BPMN:
                for (Entry<HashableGroup, SharePermission> group : this.groups.entrySet()) {
                    if (this.processDefsGroups.containsKey(group.getKey())) {
                        this.logger.trace("The default group permission is overridden by the app group permission: {}", group.getKey());
                    } else {
                        this.share(shares, model, group.getKey(), group.getValue());
                    }
                }
                for (Entry<HashableGroup, SharePermission> group : this.processDefsGroups.entrySet())
                    this.share(shares, model, group.getKey(), group.getValue());
                break;
            case Model.MODEL_TYPE_FORM:
                for (Entry<HashableGroup, SharePermission> group : this.groups.entrySet()) {
                    if (this.formDefsGroups.containsKey(group.getKey())) {
                        this.logger.trace("The default group permission is overridden by the app group permission: {}", group.getKey());
                    } else {
                        this.share(shares, model, group.getKey(), group.getValue());
                    }
                }
                for (Entry<HashableGroup, SharePermission> group : this.formDefsGroups.entrySet())
                    this.share(shares, model, group.getKey(), group.getValue());
                break;
            default:
                this.logger.trace("Unrecognized model type; not sharing: {}", model.getModelType());
        }
    }
    
    private Map<HashableGroup, SharePermission> fetchCurrentModelShares(Model model) {
        int page = 1;
        Pageable pageable = PageRequest.of(page, this.shareChunkSize);
        
        Map<HashableGroup, SharePermission> map = new HashMap<>();
        
        List<ModelShareInfo> shares = this.shareInfoRepo.findByModelIdOrderByShareDateAsc(model.getId(), pageable);
        while (!shares.isEmpty()) {
            for (ModelShareInfo share : shares)
                map.put(new HashableGroup(share.getGroup()), share.getPermission());
            
            pageable = pageable.next();
            shares = this.shareInfoRepo.findByModelIdOrderByShareDateAsc(model.getId(), pageable);
        }
        
        if (this.logger.isTraceEnabled()) {
            this.logger.trace("Found existing shares: {}", map);
        } else {
            this.logger.debug("Found existing shares: {}", map.size());
        }
        return map;
    }
    
    private void share(Map<HashableGroup, SharePermission> shares, Model model, HashableGroup group, SharePermission permission) {
        if (permission.equals(shares.get(group))) {
            this.logger.trace("The group is already shared with the same permission: {}: {}", group, permission);
        } else {
            this.share(model, group, permission);
        }
    }
    
    private void share(Model model, HashableGroup group, SharePermission permission) {
        this.logger.info("Sharing model with group: {}: {}: {}", model.getId(), model.getName(), group);
        this.shareService.shareModelWithGroup(model, group.getGroup(), permission).getShareDate();
    }
    
    public void unshareAll(Model model) {
        int page = 1;
        Pageable pageable = PageRequest.of(page, this.shareChunkSize);
        
        List<ModelShareInfo> shares = this.shareInfoRepo.findByModelIdOrderByShareDateAsc(model.getId(), pageable);
        while (!shares.isEmpty()) {
            for (ModelShareInfo share : shares)
                this.unshare(model, share);

            shares = this.shareInfoRepo.findByModelIdOrderByShareDateAsc(model.getId(), pageable);
        }
    }
    
    private void unshare(Model model, ModelShareInfo share) {
        this.logger.debug("Unsharing model from group: {}: {}: {}: {}", model.getId(), model.getName(), share.getGroup().getId(), share.getGroup().getName());
        this.shareService.deleteShareInfo(model, share.getId());
    }
    
    

    private class HashableGroup {
        
        private final Group group;
        
        private HashableGroup(Group group) {
            this.group = group;
        }
        
        public Group getGroup() {
            return this.group;
        }
        
        @Override
        public int hashCode() {
            return this.group.getId().hashCode();
        }
        
        @Override
        public boolean equals(Object obj) {
            if (obj instanceof HashableGroup) {
                return this.group.getId().equals(((HashableGroup) obj).getGroup().getId());
            } else if (obj instanceof Group) {
                return this.group.getId().equals(((Group) obj).getId());
            } else {
                return false;
            }
        }
        
        @Override
        public String toString() {
            return this.group.getId().toString() + ":" + this.group.getName();
        }
        
    }

}
